@@ -30,7 +30,7 @@ Author
 Links
 -----
 Website: http://sumudu.tennakoon.net/projects/MLToolkit
-Github: https://github.com/mltoolkit/mltk
+Github: https://mltoolkit.github.io/MLToolKit
 
 License
 -------
@@ -58,48 +58,65 @@ def plot_model_results(ResultsTable, x_column, y_column, size_column, color_colu
     bounds = np.array([0.0, 0.3, 0.5, 1.0, 2.0, 5.0, 10.0, 20.0, 25.0, 50.0])
     norm = colors.BoundaryNorm(boundaries=bounds, ncolors=256)
     ResultsTable.sort_values(by=size_column, ascending=False, inplace=True)
+    
+    max_x = max(ResultsTable[x_column])*1.05
+    max_y = max(ResultsTable[y_column])*1.05
+    
     plt.figure()
     plt.title('Model Charateristics \n ResponseFraction ~ Marker size ')
     plt.scatter(ResultsTable[x_column], ResultsTable[y_column], c=ResultsTable[color_column].values*color_scale, s=ResultsTable[size_column]*size_scale, cmap='nipy_spectral', norm=norm, marker='s')
-    plt.plot([0, 100], [0, 100], 'k:')
+    plt.plot([0, max(max_x, max_y, 1)], [0, max(max_x, max_y, 1)], 'k:')
     plt.xlabel(x_column + ' (Predicted)')
     plt.ylabel(y_column+ ' (Actual)')
     cbar = plt.colorbar()
     cbar.set_label(color_column)
     
-    plt.xlim(0, max(ResultsTable[x_column])*1.05)
-    plt.ylim(0, max(ResultsTable[y_column])*1.05)
+    plt.xlim(0, max(max_x, max_y))
+    plt.ylim(0, max(max_x, max_y))
     
 ###############################################################################
 # ROBUSTNESS TABLE                 
-def robustness_table(ResultsSet, target_variable='Response', score_variable='Probability',  score_label='Score', condensed=False, show_plot=False):  
+def robustness_table(ResultsSet, target_variable='Response', score_variable='Probability',  score_label='Score', condensed=False, problem_type='classification', formatted=True, n=None, show_plot=False):  
 
     if target_variable not in ResultsSet.columns:
         ResultsSet[target_variable]=None
         
     RobustnessTable = ResultsSet.groupby(by=[score_label]).agg({score_variable:[min,max,'mean','count'], target_variable:sum})        
     
-    RobustnessTable.columns=np.array(['min{}'.format(score_variable), 'max{}'.format(score_variable), 'mean{}'.format(score_variable), 'BucketCount', 'ResponseCount'])
+    predict_min_label = 'Min{}'.format(score_variable)
+    predict_max_label = 'Max{}'.format(score_variable)
+    predict_mean_label = 'Mean{}'.format(score_variable)
+    
+    if problem_type=='classification':
+        response_quantity_label = 'ResponseCount'
+        bucket_mean_label = 'BucketPrecision'
+        cum_bucket_mean_label = 'CumulativePrecision'
+    elif problem_type=='regression':
+        response_quantity_label = 'ResponseAmount'
+        bucket_mean_label = 'BucketMeanAmount'
+        cum_bucket_mean_label = 'CumulativeMeanAmount'
+        
+    RobustnessTable.columns=np.array([predict_min_label, predict_max_label, predict_mean_label, 'BucketCount', response_quantity_label])
     
     mean_probability = np.mean(ResultsSet[score_variable])
     min_probability = np.min(ResultsSet[score_variable])
     max_probability = np.max(ResultsSet[score_variable])
     
     total_count = np.sum(RobustnessTable['BucketCount'])
-    total_response_count = np.sum(RobustnessTable['ResponseCount'])
+    total_response_count = np.sum(RobustnessTable[response_quantity_label])
         
     RobustnessTable['BucketFraction'] = RobustnessTable['BucketCount']/total_count
-    RobustnessTable['ResponseFraction'] =RobustnessTable['ResponseCount']/total_response_count
-    RobustnessTable['BucketPrecision'] = RobustnessTable['ResponseCount']/RobustnessTable['BucketCount'] 
+    RobustnessTable['ResponseFraction'] =RobustnessTable[response_quantity_label]/total_response_count
+    RobustnessTable[bucket_mean_label] = RobustnessTable[response_quantity_label]/RobustnessTable['BucketCount'] 
     RobustnessTable['CumulativeBucketFraction'] = RobustnessTable['BucketFraction'][::-1].cumsum()
     RobustnessTable['CumulativeResponseFraction'] = RobustnessTable['ResponseFraction'][::-1].cumsum()
-    RobustnessTable['CumulativePrecision'] = RobustnessTable['ResponseCount'][::-1].cumsum()/RobustnessTable['BucketCount'][::-1].cumsum()
+    RobustnessTable[cum_bucket_mean_label] = RobustnessTable[response_quantity_label][::-1].cumsum()/RobustnessTable['BucketCount'][::-1].cumsum()
 #    RobustnessTable['CumulativeBucketCount'] = RobustnessTable['BucketCount'][::-1].cumsum()
 #    RobustnessTable['CumulativeResponseCount'] = RobustnessTable['ResponseCount'][::-1].cumsum()
     
     total_bucket_fraction = total_count/total_count
     total_response_fraction = total_response_count/total_response_count
-    mean_precision= np.sum(RobustnessTable['BucketPrecision']*RobustnessTable['BucketCount'])/total_count
+    mean_precision= np.sum(RobustnessTable[bucket_mean_label]*RobustnessTable['BucketCount'])/total_count
     
     SummaryRow = pd.DataFrame(data=[[min_probability, max_probability, mean_probability, total_count, total_response_count, total_bucket_fraction, total_response_fraction, mean_precision, total_bucket_fraction, total_response_fraction, mean_precision]], columns=RobustnessTable.columns)
     SummaryRow.index = np.array(['DataSet'])
@@ -109,16 +126,37 @@ def robustness_table(ResultsSet, target_variable='Response', score_variable='Pro
     RobustnessTable = RobustnessTable.append(SummaryRow, ignore_index=False)
     
     if show_plot==True:
-        plot_model_results(RobustnessTable[RobustnessTable.index!='DataSet'], x_column='mean{}'.format(score_variable), y_column='BucketPrecision', size_column='ResponseFraction', color_column='BucketFraction', color_scale=100, size_scale=2000)
-           
-    RobustnessTable[['BucketCount','ResponseCount']] = RobustnessTable[['BucketCount','ResponseCount']].astype('int')
+        plot_model_results(RobustnessTable[RobustnessTable.index!='DataSet'], x_column=predict_mean_label, y_column=bucket_mean_label, size_column='ResponseFraction', color_column='BucketFraction', color_scale=100, size_scale=2000)
+
+    if problem_type=='classification':
+        RobustnessTable[['BucketCount',response_quantity_label]] = RobustnessTable[['BucketCount',response_quantity_label]].astype('int')
+        if n == None: 
+            n = 4
+    elif problem_type=='regression':
+        RobustnessTable['BucketCount'] = RobustnessTable['BucketCount'].astype('int')
+        RobustnessTable['PredictedAmount'] = RobustnessTable['BucketCount'] * RobustnessTable[predict_mean_label]
+        RobustnessTable[response_quantity_label] = RobustnessTable[response_quantity_label].astype('float')
+        if n == None:
+            n = 2
+        RobustnessTable[['PredictedAmount']] = RobustnessTable[['PredictedAmount']].round(n)
     
+    if formatted:    
+        RobustnessTable[['BucketFraction', 'ResponseFraction', 'CumulativeBucketFraction', 'CumulativeResponseFraction']] = RobustnessTable[['BucketFraction', 'ResponseFraction', 'CumulativeBucketFraction', 'CumulativeResponseFraction']].round(max(n,4))
+        RobustnessTable[[predict_min_label, predict_max_label, predict_mean_label, bucket_mean_label, cum_bucket_mean_label]] = RobustnessTable[[predict_min_label, predict_max_label, predict_mean_label, bucket_mean_label, cum_bucket_mean_label]].round(n)
+
+    condensed_columns = [predict_mean_label, bucket_mean_label, 'BucketCount', 'PredictedAmount', response_quantity_label, 'CumulativeBucketFraction', 'CumulativeResponseFraction', cum_bucket_mean_label]
+    all_columns = [predict_min_label, predict_max_label, predict_mean_label, 'BucketCount', 'PredictedAmount', 
+                        response_quantity_label, 'BucketFraction', 'ResponseFraction', bucket_mean_label,
+                        'CumulativeBucketFraction', 'CumulativeResponseFraction', cum_bucket_mean_label]
+                       
     if condensed:
-        RobustnessTable = RobustnessTable[['meanProbability', 'BucketPrecision', 'BucketCount', 'ResponseCount', 'CumulativeBucketFraction',  'CumulativeResponseFraction', 'CumulativePrecision']]
+        RobustnessTable = RobustnessTable[[column for column in  condensed_columns if column in RobustnessTable.columns]]
+    else:
+        RobustnessTable = RobustnessTable[[column for column in all_columns if column in RobustnessTable.columns]]
     
     return RobustnessTable
 ###############################################################################
-def get_score_cutoffs(ResultsSet, quantiles=[0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0], target_variable='Response', score_variable='Probability'):
+def get_score_cutoffs(ResultsSet, quantiles=[0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0], target_variable='Response', problem_type='classification', score_variable='Probability'):
 
     responses = len(ResultsSet.loc[ResultsSet[target_variable]==1].index)
     total = len(ResultsSet[target_variable].index)
@@ -130,58 +168,81 @@ def get_score_cutoffs(ResultsSet, quantiles=[0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.
     edges = ResultsSet.groupby(by=['quantile'])[score_variable].min().values.round(4)
     edges = sorted(edges)
     edges = list(edges)
-    edges.append(1.0000)
+    edges.append(np.inf)
     edges[0] = 0.0  
     
-    out = pd.qcut(x=ResultsSet['rank'], q=[0.0, 1.0-response_rate, 1.0]) # cutoff to take the top most fraction matching response rate.
-    ResultsSet['quantile']=out
-    thresholds = ResultsSet.groupby(by=['quantile'])[score_variable].min().values.round(4)
-    thresholds = sorted(thresholds)
-    thresholds = list(thresholds)
-    threshold = thresholds[1]
+    if problem_type == 'classification':
+        out = pd.qcut(x=ResultsSet['rank'], q=[0.0, 1.0-response_rate, 1.0]) # cutoff to take the top most fraction matching response rate.
+        ResultsSet['quantile']=out
+        thresholds = ResultsSet.groupby(by=['quantile'])[score_variable].min().values.round(4)
+        thresholds = sorted(thresholds)
+        thresholds = list(thresholds)
+        threshold = thresholds[1]
+    else:
+        threshold = 0.0
 
     return edges, threshold
     
 ###############################################################################
 # COMPUTE MODEL PERFORMANCE EVALUATION MATRICS                  
-def model_performance_matrics(ResultsSet, target_variable='Actual', score_variable='Probability', quantile_label='Quantile',  quantiles=1000, show_plot=False):
+def model_performance_matrics(ResultsSet, target_variable='Actual', score_variable='Probability', quantile_label='Quantile',  quantiles=1000, problem_type='classification', show_plot=False):
     from sklearn import metrics #roc_curve, auc, precision_recall_curve,balanced_accuracy_score
 
     # Create quantiles
     ResultsSet[quantile_label] = pd.qcut(x=ResultsSet[score_variable], q=quantiles, labels = False, duplicates='drop')
     ResultsSet[quantile_label] = ResultsSet[quantile_label] + 1
     
-    RobustnessTable = robustness_table(ResultsSet, target_variable=target_variable, score_variable=score_variable, score_label=quantile_label, show_plot=show_plot)    
+    RobustnessTable = robustness_table(ResultsSet, target_variable=target_variable, score_variable=score_variable, score_label=quantile_label, problem_type=problem_type, show_plot=show_plot)    
     #RobustnessTable[:-1].plot(x='maxProbability', y=['CumulativePrecision', 'CumulativeBucketFraction', 'CumulativeResponseFraction'], xlim=[0.0, 1.0], ylim=[0.0, 1.05])
     #RobustnessTable[:-1].plot(x='meanProbability', y=['BucketPrecision'], xlim=[0.0, 1.0], ylim=[0.0, 1.05])
     
-    ROCCurve = {}  
-    ROCCurve['FPR'], ROCCurve['TPR'], ROCCurve['Threshold'] = metrics.roc_curve(ResultsSet[target_variable].values, ResultsSet[score_variable].values)
-    roc_auc = metrics.auc( ROCCurve['FPR'], ROCCurve['TPR'])
-    ROCCurve = pd.DataFrame(data=ROCCurve)
-    
-    # undersample curve     
-    if len(ROCCurve.index)>10000:
-        a = ROCCurve[:1]
-        b = ROCCurve[1:-1].sample(10000-2).sort_values(by='Threshold', ascending=False)
-        c = ROCCurve[-1:] 
-        ROCCurve = a.append(b).append(c).reset_index(drop=True)
-
-    PrecisionRecallCurve = {}  
-    PrecisionRecallCurve['Precision'], PrecisionRecallCurve['Recall'], PrecisionRecallCurve['Threshold'] = metrics.precision_recall_curve(ResultsSet[target_variable].values, ResultsSet[score_variable].values)
-    PrecisionRecallCurve['Threshold']=np.insert(PrecisionRecallCurve['Threshold'], 0,0)    
-    PrecisionRecallCurve = pd.DataFrame(data=PrecisionRecallCurve)
-    prc_auc = metrics.auc(PrecisionRecallCurve['Recall'], PrecisionRecallCurve['Precision'])
-    
-    # undersample curve
-    if len(PrecisionRecallCurve.index)>10000:
-        a = PrecisionRecallCurve[:1]
-        b = PrecisionRecallCurve[1:-1].sample(10000-2).sort_values(by='Threshold', ascending=True)
-        c = PrecisionRecallCurve[-1:]
-        PrecisionRecallCurve = a.append(b).append(c).reset_index(drop=True)
-    
+    if problem_type=='classification':
+        
+        # ROCCurve 
+        ROCCurve = {}  
+        ROCCurve['FPR'], ROCCurve['TPR'], ROCCurve['Threshold'] = metrics.roc_curve(ResultsSet[target_variable].values, ResultsSet[score_variable].values)
+        roc_auc = metrics.auc( ROCCurve['FPR'], ROCCurve['TPR'])
+        ROCCurve = pd.DataFrame(data=ROCCurve)
+        
+        if len(ROCCurve.index)>10000:
+            a = ROCCurve[:1]
+            b = ROCCurve[1:-1].sample(10000-2).sort_values(by='Threshold', ascending=False)
+            c = ROCCurve[-1:] 
+            ROCCurve = a.append(b).append(c).reset_index(drop=True)
+        
+        # PrecisionRecallCurve
+        PrecisionRecallCurve = {}  
+        PrecisionRecallCurve['Precision'], PrecisionRecallCurve['Recall'], PrecisionRecallCurve['Threshold'] = metrics.precision_recall_curve(ResultsSet[target_variable].values, ResultsSet[score_variable].values)
+        PrecisionRecallCurve['Threshold']=np.insert(PrecisionRecallCurve['Threshold'], 0,0)    
+        PrecisionRecallCurve = pd.DataFrame(data=PrecisionRecallCurve)
+        prc_auc = metrics.auc(PrecisionRecallCurve['Recall'], PrecisionRecallCurve['Precision'])
+        
+        if len(PrecisionRecallCurve.index)>10000:
+            a = PrecisionRecallCurve[:1]
+            b = PrecisionRecallCurve[1:-1].sample(10000-2).sort_values(by='Threshold', ascending=True)
+            c = PrecisionRecallCurve[-1:]
+            PrecisionRecallCurve = a.append(b).append(c).reset_index(drop=True)
+            
+    elif problem_type=='regression':
+        ROCCurve = None
+        PrecisionRecallCurve = None
+        roc_auc = None
+        prc_auc = None
+        
     return RobustnessTable, ROCCurve, PrecisionRecallCurve, roc_auc, prc_auc
 
+def get_regression_errors(ResultsSet, target_variable='Actual', predicted_variable='Predicted'):
+    
+    from sklearn import metrics 
+    
+    SE = (ResultsSet[target_variable] - ResultsSet[predicted_variable])**2
+    MSE = SE.mean()
+    RMSE = np.sqrt(MSE)
+    
+    R2 = metrics.r2_score(ResultsSet[target_variable], ResultsSet[predicted_variable])  
+    
+    return RMSE, R2
+    
 ###############################################################################
 #  PLOT EVALUATION MATRICS                  
 def plot_eval_matrics(RobustnessTable, ROCCurve, PrecisionRecallCurve, AUC, score_variable, figure=1, description=''):
